/*
 * sar and iostat common routines.
 * (C) 1999 by Sebastien GODARD (sebastien.godard@wanadoo.fr)
 * 
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                 *
 ***************************************************************************
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>	/* for S_IRUSR, S_IWUSR, etc. */

#ifdef USE_NLS
#include <locale.h>
#include <libintl.h>
#define _(string) gettext(string)
#else
#define _(string) (string)
#endif

extern short int sl_hour, sl_minute, sl_second;
extern long interval, file_interval;
extern unsigned long uptime[];


/*
 * Get current date or time
 */
void get_localtime(struct tm *loc_time)
{
   time_t timer;
   struct tm *ltm;
   

   time(&timer);
   ltm = localtime(&timer);

   *loc_time = *ltm;
}


/*
 * Next time stamp
 */
int next_slice(int curr)
{
   int min, max, pt1, pt2;
   

   sl_second += file_interval;
   if (sl_second >= 60) {
      sl_minute += sl_second / 60;
      sl_second %= 60;
      if (sl_minute >= 60) {
	 sl_hour += sl_minute / 60;
	 sl_minute %= 60;
	 if (sl_hour >= 24)
	    sl_hour %= 24;
      }
   }

   /*
    * A few notes about the "algorithm" used here to display selected entries
    * from the system activity file (option -f with a given interval number):
    * Let 'Iu' be the interval value given by the user on the command line,
    *     'If' the interval value used in the system activity file,
    * and 'En' the nth entry (identified by its time stamp) of the file.
    * We choose In = [ En - If/2, En + If/2 [ if If is even,
    *        or In = [ En - If/2, En + If/2 ] if not.
    * En will be displayed if
    *       (Pn * Iu) /or/ (P'n * Iu) belongs to In
    * with  Pn = En / Iu and P'n = En / Iu + 1
    */
   min = (uptime[curr] / 100) - (file_interval / 2);
   max = (uptime[curr] / 100) + (file_interval / 2) + (file_interval & 0x1);
   
   pt1 = ((uptime[curr] / 100) / interval) * interval;
   pt2 = (((uptime[curr] / 100) / interval) + 1) * interval;
   
   return (((pt1 >= min) && (pt1 < max)) || ((pt2 >= min) && (pt2 < max)));
}


/*
 * Open file for reading or writing
 */
void open_files(char from_file[], char to_file[], int *from_fd, int *to_fd)
{

   if (!from_file[0] && !to_file[0])
      return;
   
   
   if (from_file[0]) {
      if ((*from_fd = open(from_file, O_RDONLY)) < 0) {
	 fprintf(stderr, _("Cannot open %s: %s\n"), from_file, strerror(errno));
	 exit(3);
      }
   }
   else {
      if ((*to_fd = open(to_file, O_WRONLY | O_CREAT | O_TRUNC,
			 S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {
	 fprintf(stderr, _("Cannot open %s: %s\n"), to_file, strerror(errno));
	 exit(3);
      }
   }
}


/*
 * Close files
 */
void close_files(int read_file, int write_file, int from_fd, int to_fd)
{
   if (read_file)
      close(from_fd);
   if(write_file)
      close(to_fd);
}


inline void print_gal_header(char *sysname, char *release,
			     char *nodename, char *cur_date)
{
   printf("%s %s (%s) \t%s\n", sysname, release, nodename, cur_date);
}


/*
 * Find number of processors used on the machine
 * (0 means one proc, 1 means two proc, etc.)
 * As far as I know, there are two possibilities for this:
 * 1) Use /proc/stat or 2) Use /proc/cpuinfo
 * (I haven't heard of a better method to guess it...)
 */
void get_nb_proc_used(char *stat_file, int *proc_used)
{
   FILE *statfp;
   static char line[2240];
   int proc_nb;
   
   
   /* Open stat file */
   if ((statfp = fopen(stat_file, "r")) == NULL) {
      fprintf(stderr, _("Cannot open %s: %s\n"), stat_file, strerror(errno));
      exit(1);
   }

   while (fgets(line, 2240, statfp) != NULL) {
      
      if (strncmp(line, "cpu ", 4) && !strncmp(line, "cpu", 3)) {
	 sscanf(&line[3], "%d", &proc_nb);
	 if (proc_nb > *proc_used)
	   *proc_used = proc_nb;
      }
   }
   
   if (*proc_used == 0) {
      fprintf(stderr, _("Invalid %s file!\n"), stat_file);
      exit(1);
   }
   /* If proc_used < 0 then there is only one proc */
   if (*proc_used < 0)
      *proc_used = 0;

   /* Close file */
   fclose(statfp);
}
