/*
 * iostat: report I/O statistics
 * (C) 1998-1999 by Sebastien GODARD (sebastien.godard@wanadoo.fr)
 * 
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                 *
 ***************************************************************************
 */

/* Version */
#define VERSION		"2.0"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <sys/utsname.h>
#include "iostat.h"

#ifdef USE_NLS
#include <locale.h>
#include <libintl.h>
#define _(string) gettext(string)
#else
#define _(string) (string)
#endif


unsigned int cpu_user[2] = {0, 0}, cpu_nice[2] = {0, 0}, cpu_system[2] = {0, 0};
#ifdef CPU_WAIT
unsigned int cpu_wait[2] = {0, 0};
#endif
unsigned long cpu_idle[2] = {0, 0};
unsigned int dk_drive_sum[2] = {0, 0};
unsigned int dk_drive[2][NR_DISK];
unsigned int dk_drive_rblk[2][NR_DISK], dk_drive_wblk[2][NR_DISK];
#ifdef HAS_DISK_ACCT
unsigned int dk_drive_pgin[2][NR_DISK], dk_drive_pgout[2][NR_DISK];
#endif
unsigned int pgpgin[2] = {0, 0}, pgpgout[2] = {0, 0};
unsigned long uptime[2] = {0, 0};
#ifdef TTY
unsigned long tty_in[2] = {0, 0}, tty_out[2] = {0, 0};
#endif
long int interval = 0;
short int sl_hour, sl_minute, sl_second;

int proc_used = -1;	/* Nb of proc on the machine. A value of 1 means two procs... */
long file_interval;
char buffer[BUF_SIZE], timestamp[14];

char dp = '.';    /* Decimal point */


/*
 * Print usage and exit
 */
void usage(char *progname)
{
   fprintf(stderr, _("%s version %s\n"
		   "(C) S. Godard <sebastien.godard@wanadoo.fr>\n"
	           "Usage: %s [ options... ]\n"
		   "Options are:\n"
		   "[ -c | -d ] [ -p | -l ] [ -t ] [ -o <filename> | -f <filename> ]\n"
		   "[ interval [ count ] ]\n"),
	   progname, VERSION, progname);
   exit(1);
}


/* 
 * SIGALRM signal handler
 */
void alarm_handler(int sig)
{
   signal(SIGALRM, alarm_handler);
   alarm(interval);
}


/*
 * Initialize disk stat counters
 */
void init_stat(int curr)
{
   int disk_index;
   
   
   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
      dk_drive[curr][disk_index] = 0;

      dk_drive_rblk[curr][disk_index] = dk_drive_wblk[curr][disk_index] = 0;
#ifdef HAS_DISK_ACCT
      dk_drive_pgin[curr][disk_index] = dk_drive_pgout[curr][disk_index] = 0;
#endif      
   }
}


/*
 * Open iostat file for reading or writing
 */
void rw_io_header(int read_from_file, int write_to_file, int from_fd, int to_fd, struct tm *loc_time, int *long_size)
{
   int nb;
   

   if (read_from_file) {
      /* Read file header */
      nb = read(from_fd, buffer, HEADER_SIZE);
      if ((nb != HEADER_SIZE) ||
	  ((buffer[IO_OFS_MAGIC] + buffer[IO_OFS_MAGIC + 1] * 256) != IO_MAGIC)) {
      fprintf(stderr, _("Invalid iostat file\n"));
      exit(4);
      }

      /* Get date and time stored in file */
      loc_time->tm_mday = (int) buffer[IO_OFS_DAY];
      loc_time->tm_mon = (int) buffer[IO_OFS_MONTH];
      loc_time->tm_year = (int) buffer[IO_OFS_YEAR];
      sl_hour = (int) buffer[IO_OFS_HOUR];
      sl_minute = (int) buffer[IO_OFS_MIN];
      sl_second = (int) buffer[IO_OFS_SEC];
     
      /* Get interval */
      memcpy(&file_interval, &buffer[IO_OFS_INTERVAL], sizeof(long));
      
      /* System name, release number and hostname are also available in the buffer */
      
      /* Size of long saved in file */
      *long_size = buffer[IO_OFS_LONGSIZE];
	 
   }
   /* 'else' because you read OR you write... */
   else if (write_to_file){

      buffer[IO_OFS_MAGIC] = IO_MAGIC & 0x00ff;		/* iostat file magic number */
      buffer[IO_OFS_MAGIC + 1] = IO_MAGIC >> 8;

      buffer[IO_OFS_DAY] = (char) loc_time->tm_mday;	/* Store date in file */
      buffer[IO_OFS_MONTH] = (char) loc_time->tm_mon;
      buffer[IO_OFS_YEAR] = (char) loc_time->tm_year;
      
      buffer[IO_OFS_HOUR] = (char) loc_time->tm_hour;		/* Store time in file */
      buffer[IO_OFS_MIN] = (char) loc_time->tm_min;
      buffer[IO_OFS_SEC] = (char) loc_time->tm_sec;

      memcpy(&buffer[IO_OFS_INTERVAL], &interval, sizeof(long));	/* Interval */
      
      /* System name, release number and hostname already in the buffer */
      
      /* Size of long */
      buffer[IO_OFS_LONGSIZE] = sizeof(long);
      
      /* ...and write iostat file header */
      if ((nb = write(to_fd, buffer, HEADER_SIZE)) != HEADER_SIZE) {
	 fprintf(stderr, _("Cannot write iostat file header: %s\n"), strerror(errno));
	 exit(3);
      }
   }
}


/*
 * Read stats from iostat file
 */
int read_stat_from_file(int curr, int from_fd, int long_size)
{
   int nb, pos = 0;
   short disk_index;
   
   
   nb = read(from_fd, buffer, (ENTRY_LONG_NB * long_size) + (ENTRY_INT_NB * sizeof(int)));
   if (!nb)
      return 1;		/* End of iostat file */
   if (nb < 0) {
      fprintf(stderr, _("Error while reading iostat file: %s\n"), strerror(errno));
      exit(4);
   }
   else if (nb < (ENTRY_LONG_NB * long_size) + (ENTRY_INT_NB * sizeof(int))) {
      fprintf(stderr, _("End of iostat file unexpected\n"));
      exit(4);
   }
   
   uptime[curr] = 0;
   /* 
    * CPU usage already divided by the nb of procs available
    * before being written to file.
    */
   memcpy(&cpu_user[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&cpu_nice[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&cpu_system[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
#ifdef CPU_WAIT   
   memcpy(&cpu_wait[curr], &buffer[pos], sizeof(int));
   uptime[curr] = cpu_wait[curr];
#endif   
   /* Update 'pos' anyway since a entry in file must have a fixed length */
   pos += sizeof(int);
   cpu_idle[curr] = 0;
   memcpy(&cpu_idle[curr], &buffer[pos], MIN(sizeof(long), long_size));
   pos += long_size;

   /* Comput uptime now */
   uptime[curr] += cpu_user[curr] + cpu_nice[curr] + cpu_system[curr] + cpu_idle[curr];
   
#ifdef TTY   
   tty_in[curr] = 0;
   memcpy(&tty_in[curr], &buffer[pos], MIN(sizeof(long), long_size));
#endif   
   pos += long_size;
#ifdef TTY   
   tty_out[curr] = 0;
   memcpy(&tty_out[curr], &buffer[pos], MIN(sizeof(long), long_size));
#endif   
   pos += long_size;

   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
      memcpy(&dk_drive[curr][disk_index], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }

   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
      memcpy(&dk_drive_rblk[curr][disk_index], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }
   
   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
      memcpy(&dk_drive_wblk[curr][disk_index], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }

   memcpy(&pgpgin[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&pgpgout[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);

   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
#ifdef HAS_DISK_ACCT   
      memcpy(&dk_drive_pgin[curr][disk_index], &buffer[pos], sizeof(int));
#endif
      pos += sizeof(int);
   }
   
   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
#ifdef HAS_DISK_ACCT   
      memcpy(&dk_drive_pgout[curr][disk_index], &buffer[pos], sizeof(int));
#endif
      pos += sizeof(int);
   }
   
   /* Compute total number of I/O done */
   dk_drive_sum[curr] = dk_drive[curr][0] + dk_drive[curr][1] +
                        dk_drive[curr][2] + dk_drive[curr][3];
   return 0;
}


/*
 * Write stats to iostat file
 */
void write_stat_to_file(int curr, int to_fd)
{
   int pos = 0;
   int nb, disk_index;
   
   
   memcpy(&buffer[pos], &cpu_user[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &cpu_nice[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &cpu_system[curr], sizeof(int));
   pos += sizeof(int);
#ifdef CPU_WAIT   
   memcpy(&buffer[pos], &cpu_wait[curr], sizeof(int));
#endif
   /* Update 'pos' anyway since an entry in file must have a fixed length */
   pos += sizeof(int);
   memcpy(&buffer[pos], &cpu_idle[curr], sizeof(long));
   pos += sizeof(long);
   
#ifdef TTY
   memcpy(&buffer[pos], &tty_in[curr], sizeof(long));
#endif
   pos += sizeof(long);
#ifdef TTY
   memcpy(&buffer[pos], &tty_out[curr], sizeof(long));
#endif
   pos += sizeof(long);
   
   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
      memcpy(&buffer[pos], &dk_drive[curr][disk_index], sizeof(int));
      pos += sizeof(int);
   }

   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
      memcpy(&buffer[pos], &dk_drive_rblk[curr][disk_index], sizeof(int));
      pos += sizeof(int);
   }

   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
      memcpy(&buffer[pos], &dk_drive_wblk[curr][disk_index], sizeof(int));
      pos += sizeof(int);
   }

   memcpy(&buffer[pos], &pgpgin[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &pgpgout[curr], sizeof(int));
   pos += sizeof(int);

   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
#ifdef HAS_DISK_ACCT
      memcpy(&buffer[pos], &dk_drive_pgin[curr][disk_index], sizeof(int));
#endif
      pos += sizeof(int);
   }
   
   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
#ifdef HAS_DISK_ACCT
      memcpy(&buffer[pos], &dk_drive_pgout[curr][disk_index], sizeof(int));
#endif
      pos += sizeof(int);
   }

   if ((nb = write(to_fd, buffer, pos)) != pos) {
      fprintf(stderr, _("Cannot write data to iostat file: %s\n"), strerror(errno));
      exit(3);
   }
}


/*
 * Read stats...
 * (see linux source file linux/fs/proc/array.c)
 */
void read_stat(int curr)
{
   FILE *statfp;
   int sflag = 0;
   char resource_name[16];
   char line[80];
   unsigned int cc_user, cc_nice, cc_system;
   unsigned long cc_idle;
#ifdef CPU_WAIT
   unsigned int cc_wait;
#endif
   
   
   /* Open stat file */
   if ((statfp = fopen(STAT, "r")) == NULL) {
      perror("fopen");
      exit(1);
   }

#ifndef HAS_DISK_ACCT   
   while (sflag != F_MASK)
#else
   while (sflag != (F_MASK | F_DISK_PGIN | F_DISK_PGOUT))
#endif      
     {
      if (fgets(line, 80, statfp) == NULL) {
	 fprintf(stderr, _("Cannot read %s file\n"), STAT);
	 exit(1);
      }
      
      /*
       * Here the trailing space is necessary!
       * We want CPU usage among *all* processors and not for a particular one.
       */
      if (!strncmp(line, "cpu ", 4)) {
	 /*
	  * Read the number of jiffies spent in user, nice, system and idle mode
	  * and compute the uptime of the system in jiffies (1/100ths of a second)
	  */
#ifdef CPU_WAIT	 
	 sscanf(line, "%s  %u %u %u %lu %u",
	        resource_name, &cc_user, &cc_nice, &cc_system, &cc_idle, &cc_wait);
	 
	 /* Compute uptime */
	 uptime[curr] = (cc_user + cc_nice + cc_system + cc_idle + cc_wait) / (proc_used + 1);

	 cpu_wait[curr]   = cc_wait   / (proc_used + 1);
#else	 
	 sscanf(line, "%s  %u %u %u %lu",
	        resource_name, &cc_user, &cc_nice, &cc_system, &cc_idle);
	 
	 /* Compute uptime */
	 uptime[curr] = cc_user + cc_nice + cc_system + cc_idle;
#endif
	 cpu_user[curr]   = cc_user   / (proc_used + 1);
	 cpu_nice[curr]   = cc_nice   / (proc_used + 1);
	 cpu_system[curr] = cc_system / (proc_used + 1);
	 cpu_idle[curr]   = cc_idle   / (proc_used + 1);

	 sflag |= F_CPU;
      }

      else if (!strncmp(line, "disk_rblk", 9)) {
	 /*
	  * Read the number of blocks read from disk (one block is 1024 bytes).
	  * A quick glance at the linux kernel source file linux/drivers/block/ll_rw_blk.c
	  * once made me think this was a number of sectors. Yet, it seems to be
	  * a number of kilobytes... Please tell me if I'm wrong.
	  */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_rblk[curr][0], &dk_drive_rblk[curr][1],
		&dk_drive_rblk[curr][2], &dk_drive_rblk[curr][3]);
	 sflag |= F_DISK_RBLK;
      }
      
      else if (!strncmp(line, "disk_wblk", 9)) {
	 /* Read the number of blocks written to disk */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_wblk[curr][0], &dk_drive_wblk[curr][1],
		&dk_drive_wblk[curr][2], &dk_drive_wblk[curr][3]);
	 sflag |= F_DISK_WBLK;
      }

      /*
       * Note the trailing space here...
       * It is necessary if we don't want to confuse this line with other ones
       * (disk_rio or disk_wio for example)
       */
      else if (!strncmp(line, "disk ", 5)) {
	 /* Read tne number of I/O done since the last reboot */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive[curr][0], &dk_drive[curr][1],
		&dk_drive[curr][2], &dk_drive[curr][3]);
	 sflag |= F_DISK;
      }

      else if (!strncmp(line, "page", 4)) {
	 /* Read tne number of pages brought in and out */
	 sscanf(line, "%s %u %u",
		resource_name, &pgpgin[curr], &pgpgout[curr]);
	 sflag |= F_PAGE;
      }

#ifdef TTY
      else if (!strncmp(line, "tty", 3)) {
	 /* Read the number of characters read and written for all the ttys */
	 sscanf(line, "%s %lu %lu",
		resource_name, &tty_in[curr], &tty_out[curr]);
	 sflag |= F_TTY;
      }
#endif      

#ifdef HAS_DISK_ACCT
      /* The following exists only if the disk accounting patch has been applied */
      else if (!strncmp(line, "disk_pgin", 9)) {
	 /* Read the number of Kb read from disk */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_pgin[curr][0], &dk_drive_pgin[curr][1],
		&dk_drive_pgin[curr][2], &dk_drive_pgin[curr][3]);
	 sflag |= F_DISK_PGIN;
      }

      else if (!strncmp(line, "disk_pgout", 10)) {
	 /* Read the number of Kb written to disk */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_pgout[curr][0], &dk_drive_pgout[curr][1],
		&dk_drive_pgout[curr][2], &dk_drive_pgout[curr][3]);
	 sflag |= F_DISK_PGOUT;
      }
#endif      
   }

   /* Close stat file */
   fclose(statfp);

   /* Compute total number of I/O done */
   dk_drive_sum[curr] = dk_drive[curr][0] + dk_drive[curr][1] +
                        dk_drive[curr][2] + dk_drive[curr][3];
}



/*
 * Print everything now (stats and uptime)
 * Notes about the formula used to display stats as:
 * (x(t2) - x(t1)) / (t2 - t1) = XX.YY:
 * We have the identity: a = (a / b) * b + a % b   (a and b are integers).
 * Apply this with a = x(t2) - x(t1) (values about which stats are to be displayed)
 *             and b = t2 - t1 (elapsed time in seconds).
 * Since uptime is given in jiffies, it is always divided by 100 to get seconds.
 * The integer part XX is: a / b
 * The decimal part YY is: ((a % b) * 100) / b  (multiplied by 100 since we want YY and not 0.YY)
 */
int write_stat(int curr, int disp, int write_to_file, int to_fd, int read_from_file)
{
   int disk_index;
   unsigned long udec_part, ndec_part, sdec_part;
   unsigned long wdec_part = 0;
   struct tm loc_time;
   unsigned long itv;
   

   get_localtime(&loc_time);

   /* Get current time */
   if (!read_from_file)
      strftime(timestamp, 14, "%X  ", &loc_time);
   else {
      if (!next_slice(curr))
	 /* Not close enough to desired interval */
	 return 0;
      
      loc_time.tm_hour = sl_hour;
      loc_time.tm_min = sl_minute;
      loc_time.tm_sec = sl_second;
      
      strftime(timestamp, 14, "%X  ", &loc_time); 
   }

   /* Print time stamp */
   if (DISPLAY_TIMESTAMP(disp) || read_from_file)
      printf(_("Time: %s\n"), timestamp);
   
   itv = uptime[curr] - uptime[!curr];

   if (!DISPLAY_DISK_ONLY(disp)) {

#ifdef TTY      
      printf(_("tty:     tin       tout   "));
#endif
      
      printf(_("avg-cpu:  %%user   %%nice    %%sys   %%idle"));
   
#ifdef CPU_WAIT
      printf(_("  %%iowait"));
#endif

      printf("\n");
   
   
#ifdef TTY	     
      printf("   %9lu  %9lu   ",
	     (tty_in[curr] - tty_in[!curr]), (tty_out[curr] - tty_out[!curr]));
#endif	     

      udec_part = (((cpu_user[curr] - cpu_user[!curr]) % (itv / 100)) * 100)
		   / (itv / 100);
      ndec_part = (((cpu_nice[curr] - cpu_nice[!curr]) % (itv / 100)) * 100)
		   / (itv / 100);
      sdec_part = (((cpu_system[curr] - cpu_system[!curr]) % (itv / 100)) * 100)
		   / (itv / 100);
      
      printf("         %3lu%c%02lu  %3lu%c%02lu  %3lu%c%02lu",
	     (cpu_user[curr] - cpu_user[!curr]) / (itv / 100), dp, udec_part,
	     (cpu_nice[curr] - cpu_nice[!curr]) / (itv / 100), dp, ndec_part,
	     (cpu_system[curr] - cpu_system[!curr]) / (itv / 100), dp, sdec_part);
      
#ifdef CPU_WAIT
      wdec_part = (((cpu_wait[curr] - cpu_wait[!curr]) % (itv / 100)) * 100)
		   / (itv / 100);
      printf("  %3lu%c%02lu",
	     (cpu_wait[curr] - cpu_wait[!curr]) / (itv / 100), dp, wdec_part);
#endif
      
      printf("  %3lu%c%02lu",
	     (cpu_idle[curr] - cpu_idle[!curr]) / (itv / 100), dp,
	     /* Correct rounding error */
	     (400 - (udec_part + ndec_part + sdec_part + wdec_part)) % 100);
      
      printf("\n");
   }
   
   if (!DISPLAY_CPU_ONLY(disp)) {
      
      if (!DISPLAY_LOGICAL_IO(disp)) {

#ifdef HAS_DISK_ACCT	 
	 printf(_("Disks:      tps    Kb_read/s    Kb_wrtn/s    Kb_read    Kb_wrtn\n"));

	 for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
	  
	    printf(_("hdisk%d  %4lu%c%02lu     %5lu%c%02lu     %5lu%c%02lu    %7u    %7u\n"),
		   disk_index,
		   (dk_drive[curr][disk_index] - dk_drive[!curr][disk_index]) / (itv / 100), dp,
		   (((dk_drive[curr][disk_index] - dk_drive[!curr][disk_index]) % (itv / 100)) * 100)
		   / (itv / 100),
		   (dk_drive_pgin[curr][disk_index] - dk_drive_pgin[!curr][disk_index]) / (itv / 100), dp,
		   (((dk_drive_pgin[curr][disk_index] - dk_drive_pgin[!curr][disk_index]) % (itv / 100)) * 100)
		   / (itv / 100),
		   (dk_drive_pgout[curr][disk_index] - dk_drive_pgout[!curr][disk_index]) / (itv / 100), dp,
		   (((dk_drive_pgout[curr][disk_index] - dk_drive_pgout[!curr][disk_index]) % (itv / 100)) * 100)
		   / (itv / 100),
		   (dk_drive_pgin[curr][disk_index] - dk_drive_pgin[!curr][disk_index]),
		   (dk_drive_pgout[curr][disk_index] - dk_drive_pgout[!curr][disk_index]));
	 }
#else	 
	 printf(_("            tps    Kb_read/s    Kb_wrtn/s    Kb_read    Kb_wrtn\n"));
#endif	 
	 printf(_("Total:  %4lu%c%02lu     %5lu%c%02lu     %5lu%c%02lu    %7u    %7u\n\n"),
	        (dk_drive_sum[curr] - dk_drive_sum[!curr]) / (itv / 100), dp,
	        (((dk_drive_sum[curr] - dk_drive_sum[!curr]) % (itv / 100)) * 100)
		/ (itv / 100),
	        (pgpgin[curr] - pgpgin[!curr]) / (itv / 100), dp,
	        (((pgpgin[curr] - pgpgin[!curr]) % (itv / 100)) * 100)
		/ (itv / 100),
	        (pgpgout[curr] - pgpgout[!curr]) / (itv / 100), dp,
	        (((pgpgout[curr] - pgpgout[!curr]) % (itv / 100)) * 100)
		/ (itv / 100),
	        (pgpgin[curr] - pgpgin[!curr]),
	        (pgpgout[curr] - pgpgout[!curr]));
      }
      
      else {
	     
	 printf(_("Disks:      tps   Blk_read/s   Blk_wrtn/s   Blk_read   Blk_wrtn\n"));

	 for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
	  
	    printf(_("hdisk%d  %4lu%c%02lu     %5lu%c%02lu     %5lu%c%02lu    %7u    %7u\n"),
		   disk_index,
		   (dk_drive[curr][disk_index] - dk_drive[!curr][disk_index]) / (itv / 100), dp,
		   (((dk_drive[curr][disk_index] - dk_drive[!curr][disk_index]) % (itv / 100)) * 100)
		   / (itv / 100),
		   (dk_drive_rblk[curr][disk_index] - dk_drive_rblk[!curr][disk_index]) / (itv / 100), dp,
		   (((dk_drive_rblk[curr][disk_index] - dk_drive_rblk[!curr][disk_index]) % (itv / 100)) * 100)
		   / (itv / 100),
		   (dk_drive_wblk[curr][disk_index] - dk_drive_wblk[!curr][disk_index]) / (itv / 100), dp,
		   (((dk_drive_wblk[curr][disk_index] - dk_drive_wblk[!curr][disk_index]) % (itv / 100)) * 100)
		   / (itv / 100),
		   (dk_drive_rblk[curr][disk_index] - dk_drive_rblk[!curr][disk_index]),
		   (dk_drive_wblk[curr][disk_index] - dk_drive_wblk[!curr][disk_index]));
	 }
	 printf("\n");
      }
   }
   
   /* Write data to file if necessary */
   if (write_to_file)
      write_stat_to_file(curr, to_fd);
   
   return 1;
}


/*
 * Main entry to the program
 */
int main(int argc, char **argv)
{
   int it = 0, disp = 0;
   int opt = 1, curr = 1;
   int from_fd, to_fd, i, next, long_size;
   long int count = 1;
   char from_file[MAX_FILE_LEN] = {"\0"}, to_file[MAX_FILE_LEN] = {"\0"};
   char cur_date[11];
#ifdef USE_NLS
   struct lconv *lc;
#endif
   struct tm loc_time;
   struct utsname header;

   
#ifdef USE_NLS

   /* Init National Language Support */
   setlocale(LC_MESSAGES, "");
   setlocale(LC_TIME, "");
   setlocale(LC_NUMERIC, "");

   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);

   /* Get decimal point */
   lc = localeconv();
   dp = *(lc->decimal_point);

#endif   
   
   /* Init stat counters */
   init_stat(!curr);
   
   /* How many processors on this machine ? */
   get_nb_proc_used(STAT, &proc_used);

   /* Process args... */
   while (opt < argc) {

      if (!strcmp(argv[opt], "-o")) {	/* Output to specified file		*/
	 if (argv[++opt] && strncmp(argv[opt], "-", 1) &&
	     (strspn(argv[opt], DIGITS) != strlen(argv[opt])))
	   strcpy(to_file, argv[opt++]);
	 else
	   usage(argv[0]);
      }

      else if (!strcmp(argv[opt], "-f")) { /* Read from specified file		*/
	 if (argv[++opt] && strncmp(argv[opt], "-", 1) &&
	     (strspn(argv[opt], DIGITS) != strlen(argv[opt])))
	   strcpy(from_file, argv[opt++]);
	 else
	   usage(argv[0]);
      }

      else if (!strncmp(argv[opt], "-", 1)) {
	 for (i = 1; *(argv[opt] + i); i++) {
	    
	    switch (*(argv[opt] + i)) {

	     case 'c':
	       disp |= D_CPU_ONLY;	/* Display cpu usage only		*/
	       break;
	       
	     case 'd':
	       disp |= D_DISK_ONLY;	/* Display disk utilization only	*/
	       break;
	       
	     case 'l':
	       disp |= D_LOGICAL_IO;	/* Display logical I/O block stats	*/
	       break;
	       
	     case 'p':
	       disp &= ~D_LOGICAL_IO;	/* Display physical I/O block stats	*/
	       break;

	     case 't':
	       disp |= D_TIMESTAMP;	/* Display timestamp	       		*/
	       break;
	       
	     default:
	       usage(argv[0]);
	    }
	 }
	 opt++;
      }
	       
      else if (!it) {
	 interval = atol(argv[opt++]);
	 if (interval < 1)
 	   usage(argv[0]);
	 count = -1;
	 it = 1;
      }
      
      else {
	 count = atol(argv[opt++]);
	 if (count < 1)
	   usage(argv[0]);
      }
   }

   /* You read from a file OR you write to it... */
   if (from_file[0] && to_file[0])
      usage(argv[0]);
   
   /* Interval must be set if stat are read from or written to file */
   if (!interval && (from_file[0] || to_file[0])) {
      fprintf(stderr, _("Please give an interval value\n"));
      exit(2);
   }

   /* Open iostat files then read or write header */
   open_files(from_file, to_file, &from_fd, &to_fd);

   get_localtime(&loc_time);
   
   /*
    * Get system name, release number and hostname.
    * This definition will be overriden if read from io file.
    */
   uname(&header);
   strncpy(&buffer[IO_OFS_SYSNAME], header.sysname, UTSNAME_LEN);
   buffer[IO_OFS_SYSNAME + UTSNAME_LEN - 1] = '\0';
   strncpy(&buffer[IO_OFS_RELNAME], header.release, UTSNAME_LEN);
   buffer[IO_OFS_RELNAME + UTSNAME_LEN - 1] = '\0';
   strncpy(&buffer[IO_OFS_NODENAME], header.nodename, UTSNAME_LEN);
   buffer[IO_OFS_NODENAME + UTSNAME_LEN - 1] = '\0';

   /* Read or write io file header */
   rw_io_header(from_file[0], to_file[0], from_fd, to_fd, &loc_time, &long_size);

   strftime(cur_date, 11, "%x", &loc_time);
   print_gal_header(&buffer[IO_OFS_SYSNAME], &buffer[IO_OFS_RELNAME],
		    &buffer[IO_OFS_NODENAME], cur_date);
      
   /* Set a handler for SIGALRM */
   if (!from_file[0])
	 alarm_handler(0);

   /* Main loop */
   do {
      if (from_file[0]) {
	 /* Read stats from file */
	 if (read_stat_from_file(curr, from_fd, long_size))
	    /* End of iostat file: exit */
	    break;
      }
      else
	 /* Read kernel statistics */
	 read_stat(curr);

      /* Print results */
      if ((next = write_stat(curr, disp, to_file[0], to_fd, from_file[0]))
	  && (count > 0))
	 count--;
      fflush(stdout);
      
      if (count != 0) {
	 if (!from_file[0])
	    pause();

	 if (next)
	    curr ^= 1;
      }
   }
   while (count != 0);

   /* Close system activity files */
   close_files(from_file[0], to_file[0], from_fd, to_fd);

   return 0;
}

