/*
 * sar: report system activity
 * (C) 1999 by Sebastien GODARD (sebastien.godard@wanadoo.fr)
 * 
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                 *
 ***************************************************************************
 */

/* Version */
#define VERSION		"2.0"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/utsname.h>
#include "sar.h"

#ifdef USE_NLS
#include <locale.h>
#include <libintl.h>
#define _(string) gettext(string)
#else
#define _(string) (string)
#endif


int actflag = 0, sa_actflag = 0;
long interval = 0, count = 1;
long file_interval;
unsigned int irq_bitmap[(NR_IRQS / 32) + 1];
unsigned int cpu_bitmap[(NR_CPUS / 32) + 1];
unsigned char buf_irq_bm[MAX_IRQS_DIV_8], buf_cpu_bm[MAX_CPUS_DIV_8];
unsigned char stat_line[MAX_LINE_LEN], buffer[BUF_SIZE];
short int sl_hour, sl_minute, sl_second;

struct tstamp {
   int tm_sec;
   int tm_min;
   int tm_hour;
   int use;
};
struct tstamp tm_start, tm_end;  /* Used for -s and -e options */
   
int proc_used = -1;  /* Nb of processors on the machine. A value of 1 means two processors */
char dp = '.';       /* Decimal point */

/*
 * array[0]: stat at t
 * array[1]: stat at t'
 * array[2]: used to compute the average.
 */
unsigned int cpu_user[3] = {0, 0, 0}, cpu_nice[3] = {0, 0, 0}, cpu_system[3] = {0, 0, 0};
unsigned long cpu_idle[3] = {0, 0, 0};
unsigned int per_cpu_user[3][NR_CPUS], per_cpu_nice[3][NR_CPUS], per_cpu_system[3][NR_CPUS];
unsigned long per_cpu_idle[3][NR_CPUS];
unsigned int dk_drive[3] = {0, 0, 0};
unsigned int u_tmp[NR_DISK - 1];
unsigned int dk_drive_rio[3] = {0, 0, 0}, dk_drive_wio[3] = {0, 0, 0};
unsigned int dk_drive_rblk[3] = {0, 0, 0}, dk_drive_wblk[3] = {0, 0, 0};
unsigned long uptime[3] = {0, 0, 0};
unsigned int pgpgin[3] = {0, 0, 0}, pgpgout[3] = {0, 0, 0};
unsigned int pswpin[3] = {0, 0, 0}, pswpout[3] = {0, 0, 0};
unsigned int interrupts[3][NR_IRQS];
unsigned int irq_sum[3] = {0, 0, 0};
unsigned int context_swtch[3] = {0, 0, 0};
unsigned long total_forks[3] = {0, 0, 0};


/*
 * Print usage and exit
 */
void usage(char *progname)
{
   fprintf(stderr, _("%s version %s\n"
		   "(C) S. Godard <sebastien.godard@wanadoo.fr>\n"
	           "Usage: %s [ options... ]\n"
	           "Options are:\n"
	           "[ -A ] [ -b ] [ -B ] [ -c ] [ -i ] [ -u ] [ -w ] [ -W ]\n"
		   "[ -I { irq | -1 | -2 } ] [ -U { cpu | -1 } ]\n"
	           "[ { -o [ filename ] | -f [ filename ] [ -s [ hh:mm:ss ] ] } [ -e [ hh:mm:ss ] ] ]\n"
	           "[ interval [ count ] ]\n"),
	   progname, VERSION, progname);
   exit(1);
}


/* 
 * SIGALRM signal handler
 */
void alarm_handler(int sig)
{
   signal(SIGALRM, alarm_handler);
   alarm(interval);
}


/*
 * Init irqs array
 */
void init_irq_bitmap(unsigned int value)
{
   int i;
   

   for (i = 0; i <= NR_IRQS >> 5; i++)
     irq_bitmap[i] = value;
   
   for (i = 0; i <= MAX_IRQS_DIV_8; i++)
      buf_irq_bm[i] = 0;
}


/*
 * Init CPUs array
 */
void init_cpu_bitmap(unsigned int value)
{
   int i;
   

   for (i = 0; i <= NR_CPUS >> 5; i++)
     cpu_bitmap[i] = value;

   for (i = 0; i <= MAX_CPUS_DIV_8; i++)
      buf_cpu_bm[i] = 0;
}


void decode_time_stamp(char time_stamp[], struct tstamp *tse)
{
   time_stamp[2] = time_stamp[5] = '\0';
   tse->tm_sec = atoi(&(time_stamp[6]));
   tse->tm_min = atoi(&(time_stamp[3]));
   tse->tm_hour = atoi(time_stamp);
   tse->use = 1;
}
		       
		       
/*
 * Compare two time stamps
 */
int datecmp(struct tm *loc_time, struct tstamp *tse)
{
   if (loc_time->tm_hour == tse->tm_hour) {
      if (loc_time->tm_min == tse->tm_min)
	return (loc_time->tm_sec - tse->tm_sec);
      else
	return (loc_time->tm_min - tse->tm_min);
   }
   else
     return (loc_time->tm_hour - tse->tm_hour);
}


/*
 * Check if the user has the right to use -U option.
 * Note that he may uses this option when reading from a file,
 * even if its machine is not an SMP one...
 * This routine is called only if we are *not* reading from a file.
 */
void check_smp_option(void)
{
   unsigned int cpu_max, j = 0;
   int i;

   
   if (proc_used == 0) {
      fprintf(stderr, _("Not an SMP machine...\n"));
      exit(2);
   }

   cpu_max = 1 << ((proc_used + 1) & 0x1f);

   for (i = (cpu_max >> 5) + 1; i <= NR_CPUS >> 5; i++)
      j |= cpu_bitmap[i];

   if (j || (cpu_bitmap[proc_used >> 5] >= cpu_max)) {
      fprintf(stderr, _("Not that many processors!\n"));
      exit(2);
   }
}


/*
 * Initialize interrupt stat counters
 */
void init_stat_irq(void)
{
   int nr_irq, i;
   

   for (i = 0; i < 3; i++)
      for (nr_irq = 0; nr_irq < NR_IRQS; nr_irq++)
	 interrupts[i][nr_irq] = 0;
}


/*
 * Initialize cpu stat counters
 */
void init_stat_cpu(void)
{
   int nr_cpu, i;
   

   for (i = 0; i < 3; i++)
      for (nr_cpu = 0; nr_cpu < NR_CPUS; nr_cpu++) {
	 per_cpu_user[i][nr_cpu] = 0;
	 per_cpu_nice[i][nr_cpu] = 0;
	 per_cpu_system[i][nr_cpu] = 0;
	 per_cpu_idle[i][nr_cpu] = 0;
      }
}


/*
 * Save stats values (will be used to compute stats average)
 */
void trans_stat(int save, int curr)
{
   int i;
   
   
   total_forks[save] = total_forks[curr];
   context_swtch[save] = context_swtch[curr];

   cpu_user[save] = cpu_user[curr];
   cpu_nice[save] = cpu_nice[curr];
   cpu_system[save] = cpu_system[curr];
   cpu_idle[save] = cpu_idle[curr];
   for (i = 0; i < NR_CPUS; i++) {
      per_cpu_user[save][i] = per_cpu_user[curr][i];
      per_cpu_nice[save][i] = per_cpu_nice[curr][i];
      per_cpu_system[save][i] = per_cpu_system[curr][i];
      per_cpu_idle[save][i] = per_cpu_idle[curr][i];
   }
   
   irq_sum[save] = irq_sum[curr];
   for (i = 0; i < NR_IRQS; i++)
      interrupts[save][i] = interrupts[curr][i];
   
   pgpgin[save] = pgpgin[curr];
   pgpgout[save] = pgpgout[curr];
   pswpin[save] = pswpin[curr];
   pswpout[save] = pswpout[curr];
   dk_drive[save] = dk_drive[curr];
   dk_drive_rio[save] = dk_drive_rio[curr];
   dk_drive_wio[save] = dk_drive_wio[curr];
   dk_drive_rblk[save] = dk_drive_rblk[curr];
   dk_drive_wblk[save] = dk_drive_wblk[curr];
   
   uptime[save] = uptime[curr];
}


/*
 * Read stats from system activity file
 */
int read_stat_from_file(int curr, int from_fd, int entry_size, int long_size)
{
   int nb, pos = 0;
   
   
   nb = read(from_fd, buffer, entry_size);
   if (!nb)
      return 1;        /* End of system activity file */
   if (nb < 0) {
      fprintf(stderr, _("Error while reading system activity file: %s\n"), strerror(errno));
      exit(4);
   }
   else if (nb < entry_size) {
      fprintf(stderr, _("End of system activity file unexpected\n"));
      exit(4);
   }

   if (GET_FORK(sa_actflag)) {
      total_forks[curr] = 0;
      memcpy(&total_forks[curr], &buffer[pos], MIN(sizeof(long), long_size));
      pos += long_size;
   }

   if (GET_CTXSW(sa_actflag)) {
      memcpy(&context_swtch[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }

   if (GET_CPU(sa_actflag)) {
      /*
       * No need to divide by (proc_used + 1).
       * Stats are divided before being written to file...
       */
      memcpy(&cpu_user[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      memcpy(&cpu_nice[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      memcpy(&cpu_system[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      cpu_idle[curr] = 0;
      memcpy(&cpu_idle[curr], &buffer[pos], MIN(sizeof(long), long_size));
      pos += long_size;
   }

   if (GET_IRQ(sa_actflag)) {
      memcpy(&irq_sum[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }

   if (GET_PAGE(sa_actflag)) {
      memcpy(&pgpgin[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      memcpy(&pgpgout[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }

   if (GET_SWAP(sa_actflag)) {
      memcpy(&pswpin[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      memcpy(&pswpout[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }

   if (GET_IO(sa_actflag)) {
      memcpy(&dk_drive[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      memcpy(&dk_drive_rio[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      memcpy(&dk_drive_wio[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      memcpy(&dk_drive_rblk[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
      memcpy(&dk_drive_wblk[curr], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }
   
   if (GET_ONE_CPU(sa_actflag)) {
      for (nb = 0; nb < NR_CPUS; nb++) {
	 if (cpu_bitmap[nb >> 5] & (1 << (nb & 0x1f))) {
	    /* This proc is available *and* we want stats about it */
	    memcpy(&per_cpu_user[curr][nb], &buffer[pos], sizeof(int));
	    pos += sizeof(int);
	    memcpy(&per_cpu_nice[curr][nb], &buffer[pos], sizeof(int));
	    pos += sizeof(int);
	    memcpy(&per_cpu_system[curr][nb], &buffer[pos], sizeof(int));
	    pos += sizeof(int);
	    per_cpu_idle[curr][nb] = 0;
	    memcpy(&per_cpu_idle[curr][nb], &buffer[pos], MIN(sizeof(long), long_size));
	    pos += long_size;
	 }
	 else if (buf_cpu_bm[nb >> 3] & (1 << (nb & 0x07)))	/* Assmume NR_CPUS <= 256 */
	    /* This proc is available but we don't want stats about it */
	    pos += 3 * sizeof(int) + long_size;
      }
   }

   if (GET_ONE_IRQ(sa_actflag)) {
      for (nb = 0; nb < NR_IRQS; nb++) {
	 if (irq_bitmap[nb >> 5] & (1 << (nb & 0x1f))) {
	    /* This irq is available *and* we want stats about it */
	    memcpy(&interrupts[curr][nb], &buffer[pos], sizeof(int));
	    pos += sizeof(int);
	 }
	 else if (buf_irq_bm[nb >> 3] & (1 << (nb & 0x07)))    /* Assume NR_IRQS <= 256 */
	    /* This irq is available but we don't want stats about it */
	    pos += sizeof(int);
      }
   }

   return 0;
}


/*
 * Read stats...
 * (see linux source file linux/fs/proc/array.c)
 */
void read_stat(int curr)
{
   FILE *statfp;
   static char resource_name[16];
   static char line[2240];
   int i, pos;
   int proc_nb;
   unsigned int cc_user, cc_nice, cc_system;
   unsigned long cc_idle;
   
   
   /* Open stat file */
   if ((statfp = fopen(STAT, "r")) == NULL) {
      fprintf(stderr, _("Cannot open %s: %s\n"), STAT, strerror(errno));
      exit(1);
   }

   while (fgets(line, 2240, statfp) != NULL) {
      
      if (!strncmp(line, "cpu ", 4)) {
	 /* Read the number of jiffies spent in user, nice, system and idle mode among all proc */
	 sscanf(line, "%s  %u %u %u %lu",
	        resource_name, &cc_user, &cc_nice, &cc_system, &cc_idle);
	 cpu_user[curr]   = cc_user   / (proc_used + 1);
	 cpu_nice[curr]   = cc_nice   / (proc_used + 1);
	 cpu_system[curr] = cc_system / (proc_used + 1);
	 cpu_idle[curr]   = cc_idle   / (proc_used + 1);

	 /* Compute the uptime of the system in jiffies (1/100ths of a second) */
	 uptime[curr] = (cc_user + cc_nice + cc_system + cc_idle) / (proc_used + 1);
      }

      else if (!strncmp(line, "cpu", 3)) {
	 /* Read the number of jiffies spent in user, nice, system and idle mode for curr proc */
	 sscanf(&line[3], "%d %u %u %u %lu",
	        &proc_nb, &cc_user, &cc_nice, &cc_system, &cc_idle);
	 per_cpu_user[curr][proc_nb]   = cc_user;
	 per_cpu_nice[curr][proc_nb]   = cc_nice;
	 per_cpu_system[curr][proc_nb] = cc_system;
	 per_cpu_idle[curr][proc_nb]   = cc_idle;
      }

      else if (!strncmp(line, "disk ", 5)) {
	 /* Read number of I/O done since the last reboot */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }
   
      else if (!strncmp(line, "disk_rio", 8)) {
	 /* Read number of read I/O */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_rio[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive_rio[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }

      else if (!strncmp(line, "disk_wio", 8)) {
	 /* Read number of write I/O */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_wio[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive_wio[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }

      else if (!strncmp(line, "disk_rblk", 9)) {
	 /* Read number of blocks read from disk (one block is 1024 bytes) */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_rblk[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive_rblk[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }

      else if (!strncmp(line, "disk_wblk", 9)) {
	 /* Read number of blocks written to disk */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_wblk[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive_wblk[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }
      
      else if (!strncmp(line, "page", 4)) {
	 /* Read number of pages the system paged in and out */
	 sscanf(line, "%s %u %u",
		resource_name, &pgpgin[curr], &pgpgout[curr]);
      }
      
      else if (!strncmp(line, "swap", 4)) {
	 /* Read number of swap pages brought in and out */
	 sscanf(line, "%s %u %u",
		resource_name, &pswpin[curr], &pswpout[curr]);
      }

      else if (!strncmp(line, "intr", 4)) {
	 /* Read total number of interrupts received since system boot */
	 sscanf(line, "%s %u", resource_name, &irq_sum[curr]);
	 pos = strcspn(line + 5, " ") + 5;
	 
	 /* Read number of each interrupts received since system boot */
	 for (i = 0; i < NR_IRQS; i++) {
	    sscanf(line + pos, " %u", &interrupts[curr][i]);
	    pos += strcspn(line + pos + 1, " ") + 1;
	 }
      }
	 
      else if (!strncmp(line, "ctxt", 4)) {
	 /* Read number of context switches */
	 sscanf(line, "%s %u",
		resource_name, &context_swtch[curr]);
      }
      
      else if (!strncmp(line, "processes", 9)) {
	 /* Read number of processes created since system boot */
	 sscanf(line, "%s %lu",
		resource_name, &total_forks[curr]);
      }
   }

   /* Close stat file */
   fclose(statfp);
}


/*
 * Read header from system activity file
 */
void read_sa_header(int from_fd, struct tm *loc_time, char *cur_date, char *cur_time,
		    int *entry_size, int *long_size)
{
   int nb, i;
   int ib = 0, cb = 0;
   

   nb = read(from_fd, buffer, HEADER_SIZE);
   if ((nb != HEADER_SIZE) ||
       ((buffer[SA_OFS_MAGIC] + buffer[SA_OFS_MAGIC + 1] * 256) != SA_MAGIC)) {
      fprintf(stderr, _("Invalid system activity file\n"));
      exit(4);
   }


   /* Get date and time */
   loc_time->tm_mday = (int) buffer[SA_OFS_DAY];
   loc_time->tm_mon  = (int) buffer[SA_OFS_MONTH];
   loc_time->tm_year = (int) buffer[SA_OFS_YEAR];
   loc_time->tm_hour = (int) buffer[SA_OFS_HOUR];
   loc_time->tm_min  = (int) buffer[SA_OFS_MIN];
   loc_time->tm_sec  = (int) buffer[SA_OFS_SEC];
   strftime(cur_date, 11, "%x", loc_time);
   strftime(cur_time, 12, "%X", loc_time);


   /*
    * Set activity flag: display chosen activities *if* they
    * are in the system activity file
    */
   sa_actflag = buffer[SA_OFS_ACTFLAG] + buffer[SA_OFS_ACTFLAG + 1] * 256; 
   actflag &= sa_actflag;
      
   /* Set processor flag */
   for (i = 0; i <= NR_CPUS >> 5; i++) {
      cpu_bitmap[i] &= buffer[SA_OFS_PROC     + i * 4] +
                      (buffer[SA_OFS_PROC + 1 + i * 4] << 8 ) +
                      (buffer[SA_OFS_PROC + 2 + i * 4] << 16) +
                      (buffer[SA_OFS_PROC + 3 + i * 4] << 24);
      cb |= cpu_bitmap[i];
   }
   memcpy(buf_cpu_bm, &buffer[SA_OFS_PROC], MAX_CPUS_DIV_8);

   /* Set IRQ flag */
   for (i = 0; i <= NR_IRQS >> 5; i++) {
      irq_bitmap[i] &= buffer[SA_OFS_IRQ     + i * 4] +
	              (buffer[SA_OFS_IRQ + 1 + i * 4] << 8 ) +
	              (buffer[SA_OFS_IRQ + 2 + i * 4] << 16) +
	              (buffer[SA_OFS_IRQ + 3 + i * 4] << 24);
      ib |= irq_bitmap[i];
   }
   memcpy(buf_irq_bm, &buffer[SA_OFS_IRQ], MAX_IRQS_DIV_8);

   if ((actflag == 0) ||
       ((cb == 0) && (actflag == A_ONE_CPU)) ||
       ((ib == 0) && (actflag == A_ONE_IRQ)) ||
       ((cb + ib == 0) && (actflag == A_ONE_CPU + A_ONE_IRQ))) {
      fprintf(stderr, _("Requested activities not available in file\n"));
      close(from_fd);
      exit(2);
   }

   /* Get entry size */
   *entry_size = (buffer[SA_OFS_ENTSZ] + buffer[SA_OFS_ENTSZ + 1] * 256);
      
   /* Get interval */
   memcpy(&file_interval, &buffer[SA_OFS_INTERVAL], sizeof(long));

   /*
    * System name, release number and hostname are also available in the buffer
    * (offsets SA_OFS_SYSNAME, _RELNAME and _NODENAME).
    * We will use them later...
    */
   
   /* Get size of long saved in file */
   *long_size = buffer[SA_OFS_LONGSIZE];
   
   /* Set time variables */
   sl_hour   = loc_time->tm_hour;
   sl_minute = loc_time->tm_min;
   sl_second = loc_time->tm_sec;
}


/*
 * Write header to system activity file
 */
void write_sa_header(int to_fd, struct tm *loc_time, int *entry_size)
{
   int i, nb;
   int len = 0;
   

   buffer[SA_OFS_MAGIC] = SA_MAGIC & 0x00ff;	    /* System activity file magic number */
   buffer[SA_OFS_MAGIC + 1] = SA_MAGIC >> 8;
      
   buffer[SA_OFS_DAY]   = (char) loc_time->tm_mday; /* Date */
   buffer[SA_OFS_MONTH] = (char) loc_time->tm_mon;
   buffer[SA_OFS_YEAR]  = (char) loc_time->tm_year;
      
   buffer[SA_OFS_HOUR] = (char) loc_time->tm_hour;   /* Time */
   buffer[SA_OFS_MIN]  = (char) loc_time->tm_min;
   buffer[SA_OFS_SEC]  = (char) loc_time->tm_sec;
      
   buffer[SA_OFS_ACTFLAG] = actflag & 0x00ff;       /* Activity flag */
   buffer[SA_OFS_ACTFLAG + 1] = actflag >> 8;
      
   for (i = 0; i < 32; i++) {
      buffer[SA_OFS_PROC + i] = buffer[SA_OFS_IRQ + i] = 0;
   }
   for (i = 0; i <= NR_CPUS >> 5; i++) {    /* Processors */
      buffer[SA_OFS_PROC     + i * 4] = (cpu_bitmap[i] & 0x000000ff);
      buffer[SA_OFS_PROC + 1 + i * 4] = (cpu_bitmap[i] & 0x0000ff00) >> 8;
      buffer[SA_OFS_PROC + 2 + i * 4] = (cpu_bitmap[i] & 0x00ff0000) >> 16;
      buffer[SA_OFS_PROC + 3 + i * 4] = (cpu_bitmap[i] & 0xff000000) >> 24;
   }
   for (i = 0; i <= NR_IRQS >> 5; i++) {    /* IRQ */
      buffer[SA_OFS_IRQ     + i * 4] = (irq_bitmap[i] & 0x000000ff);
      buffer[SA_OFS_IRQ + 1 + i * 4] = (irq_bitmap[i] & 0x0000ff00) >> 8;
      buffer[SA_OFS_IRQ + 2 + i * 4] = (irq_bitmap[i] & 0x00ff0000) >> 16;
      buffer[SA_OFS_IRQ + 3 + i * 4] = (irq_bitmap[i] & 0xff000000) >> 24;
   }

   /* Compute entry length */
   if (GET_FORK(actflag))
     len += sizeof(long);
   if (GET_CTXSW(actflag))
     len += sizeof(int);
   if (GET_CPU(actflag))
     len += 3 * sizeof(int) + sizeof(long);
   if (GET_IRQ(actflag))
     len += sizeof(int);
   if (GET_PAGE(actflag))
     len += 2 * sizeof(int);
   if (GET_SWAP(actflag))
     len += 2 * sizeof(int);
   if (GET_IO(actflag))
     len += 5 * sizeof(int);
   if (GET_ONE_CPU(actflag)) {
      for (i = 0; i < NR_CPUS; i++) {
	 if (cpu_bitmap[i >> 5] & (1 << (i & 0x1f)))
	   len += 3 * sizeof(int) + sizeof(long);
      }
   }
   if (GET_ONE_IRQ(actflag)) {
      for (i = 0; i < NR_IRQS; i++) {
	 if (irq_bitmap[i >> 5] & (1 << (i & 0x1f)))
	   len += sizeof(int);
      }
   }
   buffer[SA_OFS_ENTSZ] = len & 0x00ff;
   buffer[SA_OFS_ENTSZ + 1] = len >> 8;
   *entry_size = len;
      
   /* Interval */
   memcpy(&buffer[SA_OFS_INTERVAL], &interval, sizeof(long));
      
   /*
    * System name, release number and hostname are already in the buffer...
    */

   /* Size of long */
   buffer[SA_OFS_LONGSIZE] = sizeof(long);
   
   if ((nb = write(to_fd, buffer, HEADER_SIZE)) != HEADER_SIZE) {
      fprintf(stderr, _("Cannot write system activity file header: %s\n"), strerror(errno));
      exit(3);
   }
}


/*
 * Print report header
 */
void print_report_hdr(int read_from_file, int from_fd, int write_to_file, int to_fd,
		      int *entry_size, int *long_size)
{
   char banner[MAX_LINE_LEN];
   int len, i;
   char cur_date[11], cur_time[12];
   struct tm loc_time;
   struct utsname header;

   
   /* Get local time */
   get_localtime(&loc_time);
   
   if (read_from_file)
      /* Read system activity file header */
      read_sa_header(from_fd, &loc_time, cur_date, cur_time, entry_size, long_size);
   else {
      strftime(cur_date, 11, "%x", &loc_time);
      strftime(cur_time, 12, "%X", &loc_time);
      
      /* Get system name, release number and hostname */
      uname(&header);
      strncpy(&buffer[SA_OFS_SYSNAME], header.sysname, UTSNAME_LEN);
      buffer[SA_OFS_SYSNAME + UTSNAME_LEN - 1] = '\0';
      strncpy(&buffer[SA_OFS_RELNAME], header.release, UTSNAME_LEN);
      buffer[SA_OFS_RELNAME + UTSNAME_LEN - 1] = '\0';
      strncpy(&buffer[SA_OFS_NODENAME], header.nodename, UTSNAME_LEN);
      buffer[SA_OFS_NODENAME + UTSNAME_LEN - 1] = '\0';
   }


   if (write_to_file)
      /* Write header file */
      write_sa_header(to_fd, &loc_time, entry_size);

   /*
    * Even if we don't use sa files, buffer contains system name,
    * release number and hostname.
    */
   print_gal_header(&buffer[SA_OFS_SYSNAME], &buffer[SA_OFS_RELNAME],
		    &buffer[SA_OFS_NODENAME], cur_date);

   sprintf(banner, "%s  ", cur_time);
   
   if (GET_FORK(actflag))
      strcat(banner, _("    fork/s"));
   
   if (GET_CTXSW(actflag))
      strcat(banner, _("   cswch/s"));
   
   if (GET_CPU(actflag))
      strcat(banner, _("     %user     %nice   %system     %idle"));

   if (GET_ONE_CPU(actflag)) {
      for (i = 0; i < NR_CPUS; i++) {
	 if (cpu_bitmap[i >> 5] & (1 << (i & 0x1f))) {
	    len = strlen(banner);
	    sprintf(banner + len, _(" %%user[%02d] %%nice[%02d] %%syst[%02d] %%idle[%02d]"), i, i, i, i);
	 }
      }
   }
   
   if (GET_IRQ(actflag))
      strcat(banner, _("    intr/s"));
   
   if (GET_ONE_IRQ(actflag)) {
      for (i = 0; i < NR_IRQS; i++) {
	 if (irq_bitmap[i >> 5] & (1 << (i & 0x1f))) {
	    len = strlen(banner);
	    sprintf(banner + len, _("  irq%03d/s"), i);
	 }
      }
   }
   
   if (GET_PAGE(actflag))
      strcat(banner, _("  pgpgin/s pgpgout/s"));
   
   if (GET_SWAP(actflag))
      strcat(banner, _("  pswpin/s pswpout/s"));
   
   if (GET_IO(actflag))
      strcat(banner, _("       tps      rtps      wtps   bread/s   bwrtn/s"));
   
   printf("\n%s\n", banner);
}


/*
 * Write stats to system activity daily file
 */
void write_stat_to_file(int curr, int to_fd)
{
   int pos = 0;
   int i;
   
   
   if (GET_FORK(actflag)) {
      memcpy(&buffer[pos], &total_forks[curr], sizeof(long));
      pos += sizeof(long);
   }
   
   if (GET_CTXSW(actflag)) {
      memcpy(&buffer[pos], &context_swtch[curr], sizeof(int));
      pos += sizeof(int);
   }
   
   if (GET_CPU(actflag)) {
      memcpy(&buffer[pos], &cpu_user[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &cpu_nice[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &cpu_system[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &cpu_idle[curr], sizeof(long));
      pos += sizeof(long);
   }
   
   if (GET_IRQ(actflag)) {
      memcpy(&buffer[pos], &irq_sum[curr], sizeof(int));
      pos += sizeof(int);
   }

   if (GET_PAGE(actflag)) {
      memcpy(&buffer[pos], &pgpgin[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &pgpgout[curr], sizeof(int));
      pos += sizeof(int);
   }
   
   if (GET_SWAP(actflag)) {
      memcpy(&buffer[pos], &pswpin[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &pswpout[curr], sizeof(int));
      pos += sizeof(int);
   }
   
   if (GET_IO(actflag)) {
      memcpy(&buffer[pos], &dk_drive[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &dk_drive_rio[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &dk_drive_wio[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &dk_drive_rblk[curr], sizeof(int));
      pos += sizeof(int);
      memcpy(&buffer[pos], &dk_drive_wblk[curr], sizeof(int));
      pos += sizeof(int);
   }
   
   if (GET_ONE_CPU(actflag)) {
      for (i = 0; i < NR_CPUS; i++) {
	 if (cpu_bitmap[i >> 5] & (1 << (i & 0x1f))) {
	    memcpy(&buffer[pos], &per_cpu_user[curr][i], sizeof(int));
	    pos += sizeof(int);
	    memcpy(&buffer[pos], &per_cpu_nice[curr][i], sizeof(int));
	    pos += sizeof(int);
	    memcpy(&buffer[pos], &per_cpu_system[curr][i], sizeof(int));
	    pos += sizeof(int);
	    memcpy(&buffer[pos], &per_cpu_idle[curr][i], sizeof(long));
	    pos += sizeof(long);
	 }
      }
   }

   if (GET_ONE_IRQ(actflag)) {
      for (i = 0; i < NR_IRQS; i++) {
	 if (irq_bitmap[i >> 5] & (1 << (i & 0x1f))) {
	    memcpy(&buffer[pos], &interrupts[curr][i], sizeof(int));
	    pos += sizeof(int);
	 }
      }
   }

   if ((i = write(to_fd, buffer, pos)) != pos) {
      fprintf(stderr, _("Cannot write data to system activity file: %s\n"), strerror(errno));
      exit(3);
   }
}


/*
 * Print everything now (stats and uptime)
 */
int write_stat(int curr, int write_to_file, int to_fd, int read_from_file,
	       int use_tm_start, int use_tm_end)
{
   int len, i;
   unsigned long udec_part, ndec_part, sdec_part;
   struct tm loc_time;
   unsigned long itv;
   

   get_localtime(&loc_time);

   /* Get current time */
   if (!read_from_file)
      len = strftime(stat_line, 14, "%X  ", &loc_time);
   else {
      if (!next_slice(curr))
	 /* Not close enough to desired interval */
	 return 0;
      
      loc_time.tm_hour = sl_hour;
      loc_time.tm_min = sl_minute;
      loc_time.tm_sec = sl_second;
      
      len = strftime(stat_line, 14, "%X  ", &loc_time); 
   }

   /* Check if time is right */
   if (use_tm_start && (datecmp(&loc_time, &tm_start) < 0))
      /* It's too soon... */
      return 0;
   
   /*
    * Theoretically, we should end with the interval value in uptime[!curr].
    * I think this is not necessarily true if the machine is heavily loaded,
    * hence the reckoning...
    */
   itv = uptime[curr] - uptime[!curr];

   /* Print number of processes created per second */
   if (GET_FORK(actflag))
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (total_forks[curr] - total_forks[!curr]) / (itv / 100), dp,
		     (((total_forks[curr] - total_forks[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));
   
   /* Print number of context switches per second */
   if (GET_CTXSW(actflag))
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (context_swtch[curr] - context_swtch[!curr]) / (itv / 100), dp,
		     (((context_swtch[curr] - context_swtch[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));
   
   /* Print CPU usage */
   if (GET_CPU(actflag)) {

      udec_part = (((cpu_user[curr] - cpu_user[!curr]) % (itv / 100)) * 100) / (itv / 100);
      ndec_part = (((cpu_nice[curr] - cpu_nice[!curr]) % (itv / 100)) * 100) / (itv / 100);
      sdec_part = (((cpu_system[curr] - cpu_system[!curr]) % (itv / 100)) * 100) / (itv / 100);
      
      len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
		     (cpu_user[curr] - cpu_user[!curr]) / (itv / 100), dp, udec_part);
      
      len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
		     (cpu_nice[curr] - cpu_nice[!curr]) / (itv / 100), dp, ndec_part);

      len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
		     (cpu_system[curr] - cpu_system[!curr]) / (itv / 100), dp, sdec_part);

      len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
		     (cpu_idle[curr] - cpu_idle[!curr]) / (itv / 100), dp,
		     /* Correct rounding error */
		     (300 - (udec_part + ndec_part + sdec_part)) % 100);
   }
   
   if (GET_ONE_CPU(actflag)) {
      for (i = 0; i < NR_CPUS; i++) {
	 if (cpu_bitmap[i >> 5] & (1 << (i & 0x1f))) {
	    
	    udec_part = (((per_cpu_user[curr][i] - per_cpu_user[!curr][i]) % (itv / 100)) * 100)
			/ (itv / 100);
	    ndec_part = (((per_cpu_nice[curr][i] - per_cpu_nice[!curr][i]) % (itv / 100)) * 100)
			/ (itv / 100);
	    sdec_part = (((per_cpu_system[curr][i] - per_cpu_system[!curr][i]) % (itv / 100)) * 100)
			/ (itv / 100);

	    len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
			   (per_cpu_user[curr][i] - per_cpu_user[!curr][i]) / (itv / 100),
			   dp, udec_part);
      
	    len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
			   (per_cpu_nice[curr][i] - per_cpu_nice[!curr][i]) / (itv / 100),
			   dp, ndec_part);

	    len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
			   (per_cpu_system[curr][i] - per_cpu_system[!curr][i]) / (itv / 100),
			   dp, sdec_part);

	    len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
			   (per_cpu_idle[curr][i] - per_cpu_idle[!curr][i]) / (itv / 100),
			   dp,
			   /* Correct rounding error */
			   (300 - (udec_part + ndec_part + sdec_part)) % 100);
	 }
      }
   }
   
   /* Print number of interrupts per second */
   if (GET_IRQ(actflag)) {
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (irq_sum[curr] - irq_sum[!curr]) / (itv / 100), dp,
		     (((irq_sum[curr] - irq_sum[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));
   }
   
   if (GET_ONE_IRQ(actflag)) {
      for (i = 0; i < NR_IRQS; i++) {
	 if (irq_bitmap[i >> 5] & (1 << (i & 0x1f)))
	    len += sprintf(stat_line + len, " %6lu%c%02lu",
			   (interrupts[curr][i] - interrupts[!curr][i]) / (itv / 100), dp,
			   (((interrupts[curr][i] - interrupts[!curr][i]) % (itv / 100)) * 100)
			   / (itv / 100));
      }
   }
   
   /* Print number of pages the system paged in and out */
   if (GET_PAGE(actflag)) {
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (pgpgin[curr] - pgpgin[!curr]) / (itv / 100), dp,
		     (((pgpgin[curr] - pgpgin[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (pgpgout[curr] - pgpgout[!curr]) / (itv / 100), dp,
		     (((pgpgout[curr] - pgpgout[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));
   }

   /* Print number of swap pages brought in and out */
   if (GET_SWAP(actflag)) {
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (pswpin[curr] - pswpin[!curr]) / (itv / 100), dp,
		     (((pswpin[curr] - pswpin[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (pswpout[curr] - pswpout[!curr]) / (itv / 100), dp,
		     (((pswpout[curr] - pswpout[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));
   }
   
   /* Print I/O stats (no distinction made between disks) */
   if (GET_IO(actflag)) {
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive[curr] - dk_drive[!curr]) / (itv / 100), dp,
		     (((dk_drive[curr] - dk_drive[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive_rio[curr] - dk_drive_rio[!curr]) / (itv / 100), dp,
		     (((dk_drive_rio[curr] - dk_drive_rio[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive_wio[curr] - dk_drive_wio[!curr]) / (itv / 100), dp,
		     (((dk_drive_wio[curr] - dk_drive_wio[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive_rblk[curr] - dk_drive_rblk[!curr]) / (itv / 100), dp,
		     (((dk_drive_rblk[curr] - dk_drive_rblk[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive_wblk[curr] - dk_drive_wblk[!curr]) / (itv / 100), dp,
		     (((dk_drive_wblk[curr] - dk_drive_wblk[!curr]) % (itv / 100)) * 100)
		     / (itv / 100));
   }

   printf("%s\n", stat_line);
   
   /* Write data to system activity file if necessary */
   if (write_to_file)
      write_stat_to_file(curr, to_fd);

   /* Check if time is right */
   if (use_tm_end && (datecmp(&loc_time, &tm_end) >= 0)) {
      /* It's too late now... */
      count = 0;
      return 0;
   }
   
   return 1;
}

/*
 * Print stats average
 */
void write_stat_avg(int curr)
{
   int len, i;
   unsigned long itv;

   
   strcpy(stat_line, _("Average:  "));
   len = strlen(stat_line);
   
   itv = uptime[curr] - uptime[2];
   if (!itv) {
      /* 
       * Aiee: null interval... This should only happen when reading stats
       * from a system activity file with a far too big interval value...
       */
      fprintf(stderr, _("Please give a smaller interval value\n"));
      exit(1);
   }
   
   if (GET_FORK(actflag))
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (total_forks[curr] - total_forks[2]) / (itv / 100), dp,
		     (((total_forks[curr] - total_forks[2]) % (itv / 100)) * 100)
		     / (itv / 100));
   
   if (GET_CTXSW(actflag))
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (context_swtch[curr] - context_swtch[2]) / (itv / 100), dp,
		     (((context_swtch[curr] - context_swtch[2]) % (itv / 100)) * 100)
		     / (itv / 100));
   
   if (GET_CPU(actflag)) {
      len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
		     (cpu_user[curr] - cpu_user[2]) / (itv / 100), dp,
		     (((cpu_user[curr] - cpu_user[2]) % (itv / 100)) * 100)
		     / (itv / 100));
      
      len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
		     (cpu_nice[curr] - cpu_nice[2]) / (itv / 100), dp,
		     (((cpu_nice[curr] - cpu_nice[2]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
		     (cpu_system[curr] - cpu_system[2]) / (itv / 100), dp,
		     (((cpu_system[curr] - cpu_system[2]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
		     (cpu_idle[curr] - cpu_idle[2]) / (itv / 100), dp,
		     (((cpu_idle[curr] - cpu_idle[2]) % (itv / 100)) * 100)
		     / (itv / 100));
   }
   
   if (GET_ONE_CPU(actflag)) {
      for (i = 0; i < NR_CPUS; i++) {
	 if (cpu_bitmap[i >> 5] & (1 << (i & 0x1f))) {
	    
	    len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
			   (per_cpu_user[curr][i] - per_cpu_user[2][i]) / (itv / 100),
			   dp,
			   (((per_cpu_user[curr][i] - per_cpu_user[2][i]) % (itv / 100)) * 100)
			   / (itv / 100));
      
	    len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
			   (per_cpu_nice[curr][i] - per_cpu_nice[2][i]) / (itv / 100),
			   dp,
			   (((per_cpu_nice[curr][i] - per_cpu_nice[2][i]) % (itv / 100)) * 100)
			   / (itv / 100));

	    len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
			   (per_cpu_system[curr][i] - per_cpu_system[2][i]) / (itv / 100),
			   dp,
			   (((per_cpu_system[curr][i] - per_cpu_system[2][i]) % (itv / 100)) * 100)
			   / (itv / 100));

	    len += sprintf(stat_line + len, "   %3lu%c%02lu%%",
			   (per_cpu_idle[curr][i] - per_cpu_idle[2][i]) / (itv / 100),
			   dp,
			   (((per_cpu_idle[curr][i] - per_cpu_idle[2][i]) % (itv / 100)) * 100)
			   / (itv / 100));
	 }
      }
   }

   if (GET_IRQ(actflag)) {
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (irq_sum[curr] - irq_sum[2]) / (itv / 100), dp,
		     (((irq_sum[curr] - irq_sum[2]) % (itv / 100)) * 100)
		     / (itv / 100));
   }

   if (GET_ONE_IRQ(actflag)) {
      for (i = 0; i < NR_IRQS; i++) {
	 if (irq_bitmap[i >> 5] & (1 << (i & 0x1f)))
	    len += sprintf(stat_line + len, " %6lu%c%02lu",
			   (interrupts[curr][i] - interrupts[2][i]) / (itv / 100), dp,
			   (((interrupts[curr][i] - interrupts[2][i]) % (itv / 100)) * 100)
			   / (itv / 100));
      }
   }
   
   if (GET_PAGE(actflag)) {
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (pgpgin[curr] - pgpgin[2]) / (itv / 100), dp,
		     (((pgpgin[curr] - pgpgin[2]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (pgpgout[curr] - pgpgout[2]) / (itv / 100), dp,
		     (((pgpgout[curr] - pgpgout[2]) % (itv / 100)) * 100)
		     / (itv / 100));
   }
   
   if (GET_SWAP(actflag)) {
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (pswpin[curr] - pswpin[2]) / (itv / 100), dp,
		     (((pswpin[curr] - pswpin[2]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (pswpout[curr] - pswpout[2]) / (itv / 100), dp,
		     (((pswpout[curr] - pswpout[2]) % (itv / 100)) * 100)
		     / (itv / 100));
   }
   
   if (GET_IO(actflag)) {
      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive[curr] - dk_drive[2]) / (itv / 100), dp,
		     (((dk_drive[curr] - dk_drive[2]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive_rio[curr] - dk_drive_rio[2]) / (itv / 100), dp,
		     (((dk_drive_rio[curr] - dk_drive_rio[2]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive_wio[curr] - dk_drive_wio[2]) / (itv / 100), dp,
		     (((dk_drive_wio[curr] - dk_drive_wio[2]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive_rblk[curr] - dk_drive_rblk[2]) / (itv / 100), dp,
		     (((dk_drive_rblk[curr] - dk_drive_rblk[2]) % (itv / 100)) * 100)
		     / (itv / 100));

      len += sprintf(stat_line + len, " %6lu%c%02lu",
		     (dk_drive_wblk[curr] - dk_drive_wblk[2]) / (itv / 100), dp,
		     (((dk_drive_wblk[curr] - dk_drive_wblk[2]) % (itv / 100)) * 100)
		     / (itv / 100));
   }
   
   printf("\n%s\n", stat_line);
}


/*
 * Main entry to the program
 */
int main(int argc, char **argv)
{
   int curr = 1, opt = 1;
   int entry_size, long_size, want_all_proc = 0;
   char time_stamp[9];
   int i, from_fd, to_fd, next = 1;
   char from_file[MAX_FILE_LEN] = {"\0"};
   char to_file[MAX_FILE_LEN] = {"\0"}, new_to_file[MAX_FILE_LEN] = {"\0"};
   int itv = 0;
   int std_log_file = 0;
   struct tm loc_time;
#ifdef USE_NLS
   struct lconv *lc;
#endif
   

#ifdef USE_NLS
   /* Init National Language Support */
   setlocale(LC_MESSAGES, "");
   setlocale(LC_TIME, "");
   setlocale(LC_NUMERIC, "");

   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);

   /* Get decimal point */
   lc = localeconv();
   dp = *(lc->decimal_point);
#endif   

   if (argc == 1)
      usage(argv[0]);

   tm_start.use = tm_end.use = 0;
   init_irq_bitmap(0);
   init_cpu_bitmap(0);

   /* How many processors on this machine ? */
   get_nb_proc_used(STAT, &proc_used);
   
   /* Process options */
   while (opt < argc) {

      if (!strcmp(argv[opt], "-I")) {
	 if (argv[++opt]) {
	    actflag |= A_ONE_IRQ;
	    /*
	     * Get irq number.
	     */
	    i = atol(argv[opt++]);
	    if ((i < -2) || (i >= NR_IRQS))
	       usage(argv[0]);
	    if (i == -2)
	       /* Set every bit */
	       init_irq_bitmap(0xffffffff);
	    else if (i == -1)
	       /* Set bit for the first 16 irq */
	       irq_bitmap[0] = 0x0000ffff;
	    else
	       irq_bitmap[i >> 5] |= 1 << (i & 0x1f);
	 }
	 else
	    usage(argv[0]);
      }

      else if (!strcmp(argv[opt], "-U")) {
	 if (argv[++opt]) {
	    actflag |= A_ONE_CPU;
	    i = atol(argv[opt++]);	/* Get cpu number */
	    /* Assume NR_CPUS <= 256... */
	    if ((i < -1) || (i >= NR_CPUS))
	      usage(argv[0]);
	    if (i == -1) {
	       /*
		* Set bit for every processor.
		* We still don't know if we are going to read stats
		* from a file or not...
		*/
	       init_cpu_bitmap(0xffffffff);
	       want_all_proc = 1;
	    }
	    else
	       cpu_bitmap[i >> 5] |= 1 << (i & 0x1f);
	 }
	 else
	    usage(argv[0]);
      }

      else if (!strcmp(argv[opt], "-o")) {
	 /* Save stats to a file */
	 if ((argv[++opt]) && strncmp(argv[opt], "-", 1) &&
	     (strspn(argv[opt], DIGITS) != strlen(argv[opt])))
	    strcpy(to_file, argv[opt++]);
	 else {
	    get_localtime(&loc_time);
	    sprintf(to_file, "%s/sa%02d", SA_DIR, loc_time.tm_mday);
	    std_log_file = 1;
	 }
      }

      else if (!strcmp(argv[opt], "-f")) {
	 /* Read stats from a file */
	 if ((argv[++opt]) && strncmp(argv[opt], "-", 1) &&
	     (strspn(argv[opt], DIGITS) != strlen(argv[opt])))
	    strcpy(from_file, argv[opt++]);
	 else {
	    get_localtime(&loc_time);
	    sprintf(from_file, "%s/sa%02d", SA_DIR, loc_time.tm_mday);
	 }
      }

      else if (!strcmp(argv[opt], "-s")) {
	 /* Get time start */
	 if ((argv[++opt]) && (strlen(argv[opt]) == 8))
	    strcpy(time_stamp, argv[opt++]);
	 else
	    strcpy(time_stamp, DEF_TMSTART);
	 decode_time_stamp(time_stamp, &tm_start);
      }

      else if (!strcmp(argv[opt], "-e")) {
	 /* Get time end */
	 if ((argv[++opt]) && (strlen(argv[opt]) == 8))
	    strcpy(time_stamp, argv[opt++]);
	 else
	    strcpy(time_stamp, DEF_TMEND);
	 decode_time_stamp(time_stamp, &tm_end);
      }

      else if (!strncmp(argv[opt], "-", 1) && strcmp(argv[opt], "-1")) {
	 /* Other options not previously tested */
	 for (i = 1; *(argv[opt] + i); i++) {

	    switch (*(argv[opt] + i)) {
	       
	     case 'A':
	       actflag |= A_FORK + A_PAGE + A_IRQ + A_IO + A_CPU + A_CTXSW + A_SWAP;
	       break;
	     case 'B':
	       actflag |= A_PAGE;
	       break;
	     case 'b':
	       actflag |= A_IO;
	       break;
	     case 'c':
	       actflag |= A_FORK;
	       break;
	     case 'i':
	       actflag |= A_IRQ;
	       break;
	     case 'u':
	       actflag |= A_CPU;
	       break;
	     case 'w':
	       actflag |= A_CTXSW;
	       break;
	     case 'W':
	       actflag |= A_SWAP;
	       break;
	     default:
	       usage(argv[0]);
	    }
	 }
	 opt++;
      }

      else if (!itv) { 				/* Get interval */
	 if (strspn(argv[opt], DIGITS) != strlen(argv[opt]))
	    usage(argv[0]);
	 interval = atol(argv[opt++]);
	 if (interval < 1)
	   usage(argv[0]);
	 count = -1;
	 itv = 1;
      }

      else {					/* Get count value */
	 if (strspn(argv[opt], DIGITS) != strlen(argv[opt]))
	    usage(argv[0]);
	 count = atol(argv[opt++]);
	 if (count < 1)
	   usage(argv[0]);
      }
   }

   /* Default is CPU activity... */
   if (!actflag) {
      if (from_file[0]) {
	 init_cpu_bitmap(0xffffffff);
	 init_irq_bitmap(0xffffffff);
	 actflag = 0xffffffff;
      }
      else
	actflag |= A_CPU;
   }
   
   /*
    * Check option dependencies
    */
   /* Use time start only when reading from a file */
   if (tm_start.use && !from_file[0]) {
      fprintf(stderr, _("Not reading from a system activity file (use -f option)\n"));
      usage(argv[0]);
   }
   /* Use time end only when reading from or writing to a file */
   if (tm_end.use && !from_file[0] && !to_file[0]) {
      fprintf(stderr, _("Not using a system activity file (use -f or -o option)\n"));
      usage(argv[0]);
   }
   /* You read from a file OR you write to it... */
   if (from_file[0] && to_file[0])
      usage(argv[0]);
   
   /* Interval must be set if stat are read from or written to file */
   if (!interval && (from_file[0] || to_file[0])) {
      fprintf(stderr, _("Please give an interval value\n"));
      exit(2);
   }
   
   /* Check the use of option -U */
   if (GET_ONE_CPU(actflag) && !from_file[0]) {
      if (want_all_proc)
	 for (i = proc_used + 1; i < ((NR_CPUS / 32) + 1) * 32; i++)
	    /* Reset every bit for proc > proc_used */
	    cpu_bitmap[i >> 5] &= ~(1 << (i & 0x1f));
      check_smp_option();
   }

   /* Open system activity files */
   open_files(from_file, to_file, &from_fd, &to_fd);
   
   init_stat_irq();
   init_stat_cpu();

   /* Print system report header */
   print_report_hdr(from_file[0], from_fd, to_file[0], to_fd, &entry_size, &long_size);
   
   if (from_file[0]) {      /* Read stats from file */
      uptime[!curr] = 0;
      if (read_stat_from_file(!curr, from_fd, entry_size, long_size)) {
	 fprintf(stderr, _("No entry found in system activity file!\n"));
	 exit(2);
      }
   }
   else {
      alarm_handler(0);     /* Set a handler for SIGALRM */
      read_stat(!curr);
   }
   
   trans_stat(2, !curr);

   if (interval) {
      if (to_file[0])
	 write_stat_to_file(!curr, to_fd);
      if (!from_file[0])
	 pause();
   }
   else {
      /* We don't use a system activity file, since interval == 0 */
      write_stat(!curr, 0, to_fd, from_file[0], 0, 0);
      exit(0);
   }
   
   /* Main loop */
   do {
      if (from_file[0]) {

	 /* Read stats from file */
	 if (read_stat_from_file(curr, from_fd, entry_size, long_size)) {
	    /* End of system activity file: display average stats now */
	    curr ^= 1;
	    break;
	 }
	 /*
	  * Unlike iostat file, the system activity file does not necessarily
	  * contains CPU usage. Thus we have to compute uptime this way when
	  * reading stats from file.
	  */
	 if (next)
	    uptime[curr] = uptime[!curr] + file_interval * 100;
	 else
	    uptime[curr] = uptime[curr] + file_interval * 100;
      }
      else
	 read_stat(curr);
      
      /* Rotate activity file if necessary */
      if (std_log_file) {
	 /* The user just specified the -o option without a filename */
	 get_localtime(&loc_time);
	 sprintf(new_to_file, "%s/sa%02d", SA_DIR, loc_time.tm_mday);
	 
	 if (strcmp(to_file, new_to_file)) {
	    
	    close_files(from_file[0], to_file[0], from_fd, to_fd);
	    strcpy(to_file, new_to_file);
	    open_files(from_file, to_file, &from_fd, &to_fd);
	    
	    /* Write new file header */
	    write_sa_header(to_fd, &loc_time, &entry_size);
	 }
      }

      /* Print results */
      if ((next = write_stat(curr, to_file[0], to_fd, from_file[0], tm_start.use, tm_end.use))
	  && (count > 0))
	 count--;
      fflush(stdout);
      
      if (count != 0) {
	 if (!from_file[0])
	    pause();

	 /*
	  * 'next' is set when stats are read from system activity file.
	  * It is set to 1 when we were close enough to desired interval.
	  * When not using sa file, its value is always 1.
	  */
	 if (next)
	    curr ^= 1;
      }
   }
   while (count != 0);
   
   /* Close system activity files */
   close_files(from_file[0], to_file[0], from_fd, to_fd);

   /* Print stat average */
   write_stat_avg(curr);
   
   return 0;
}

